"""
engine.py

The engine for our GGJ2011 game.
"""

import random, sys

import waves, objects

TOOLS = [
  objects.Wall,
  objects.LightTurret,
  objects.HeavyTurret,
  objects.BurstTurret,
  objects.TeslaTurret,
  objects.MagnetTurret,
  None,
  None, #"upgrade",
  None,
]

SIZE = (21, 18)
HEART = (SIZE[0]//2, SIZE[1]//2)
ENEMY_PHASE = 5
WAVELENGTH = 15
MAX_ENERGY = 99

EDGES = [
  (x, ns) for x in range(SIZE[0]) for ns in (0, SIZE[1] - 1)
] + [
  (ew, y) for y in range(SIZE[1]) for ew in (0, SIZE[0] - 1)
]

def zone(x, y):
  """
  Splits the board into zones "close", "middle" , "far", and "outside" based
  on distance from the heart.
  """
  if x < 0 or x >= SIZE[0] or y < 0 or y >= SIZE[1]:
    return "outside"
  elif 0 <= x < 2 or 0 <= y < 2 \
  or 0 <= SIZE[0] - x - 1 < 2 or 0 <= SIZE[1] - y - 1 < 2:
    return "far"
  elif abs(HEART[0] - x) <= 2 and abs(HEART[1] - y) <= 2:
    return "close"
  return "middle"

class State:
  """
  A State represents the entire game state. It keeps track of buildings, the
  gatherer's location and state, enemies, loot, and global resources like
  upgrades and resources.
  """
  size = SIZE
  things = None
  stuff = None
  commands = None

  def __init__(self, things=None, stuff=None, commands=None):
    self.stuff = {
      "quit": False,
      "energy": 20,
      "g_carrying": 0,
      "g_capacity": 10,
      "g_sap_rate": 2,
      "upgrade_levels": {"guy": 0, "base": 0},
      "upgrades": set(),
      "waves_inbound": [waves.get(0), waves.get(1)],
      "current_wave": 0,
      "guy_pos": None,
      "phase": 0,
      "active_tool": None,
      "detonations": [],
    }
    self.stuff.update(stuff or {})
    self.things = things or [
      [
        {"thing": None, "loot": 0}
        for y in range(self.size[1])
      ]
      for x in range(self.size[0])
    ]
    if self.things[HEART[0]][HEART[1]]["thing"] == None:
      self.things[HEART[0]][HEART[1]]["thing"] = objects.Heart()
    if self.things[HEART[0]][HEART[1] - 1]["thing"] == None:
      self.things[HEART[0]][HEART[1] - 1]["thing"] = objects.Guy()
      self.stuff["guy_pos"] = (HEART[0], HEART[1] - 1)
    try:
      for col in self.things:
        for cell in col:
          assert("thing" in cell and type(cell["loot"]) == int)
    except:
      print("Invalid things value in state setup.", sys.stderr)
      exit(1)
    self.commands = commands or []

  def copy(self):
    """
    Returns a new State that's a copy of this one.
    """
    return State(self.things, self.stuff, self.commands)

  def get(self, x, y):
    """
    Returns the cell at the given position.
    """
    return self.things[x][y]

  def guyspec(self):
    """
    Returns the position of the guy, the cell that he's in, and the guy
    himself.
    """
    pos = self.stuff["guy_pos"]
    if not pos:
      raise ValueError("The guy is gone!")
    cell = self.things[pos[0]][pos[1]]
    if not isinstance(cell["thing"], objects.Guy):
      raise ValueError("The guy went missing!")
    return (pos, cell, cell["thing"])

  def upgrade(self, track):
    """
    Advances to the next upgrade level along the given track ("guy" or "base").
    """
    newlevel = self.stuff["upgrade_levels"][track] + 1
    self.stuff["upgrade_levels"][track] = newlevel
    purchased = UPGRADES[track][newlevel]
    self.stuff["upgrades"].add(purchased)
    # TODO: HERE

  def attack(self, pos, damage, drop_loot=True):
    """
    Implements attacking things.
    """
    cell = self.get(*pos)
    thing = cell["thing"]
    if thing and hasattr(thing, "health")and thing.health != None:
      if thing.health > 0:
        thing.health -= damage
      if thing.health <= 0:
        damage += thing.health
        cell["thing"] = None
        if drop_loot:
          cell["loot"] += thing.loot_value
        if isinstance(thing, objects.BombBot) and drop_loot:
          self.detonate(pos, thing.damage)
      return damage
    return 0

  def detonate(self, pos, damage):
    """
    Implements detonations.
    """
    self.stuff["detonations"].append(pos)
    cell = self.get(*pos)
    thing = cell["thing"]
    for x in (pos[0] - 1, pos[0], pos[0] + 1):
      for y in (pos[1] - 1, pos[1], pos[1] + 1):
        if zone(x, y) != "outside":
          self.attack((x, y), damage)
          if isinstance(self.get(x, y)["thing"], objects.Guy):
            self.stuff["quit"] = True

  def advance(self):
    """
    Updates the State by advancing it a tick. Returns a list of shots that were
    fired.
    """
    # Increment the phase:
    self.stuff["phase"] = (self.stuff["phase"] + 1) % WAVELENGTH

    # First handle pending commands one by one:
    moved = False
    pos, cell, guy = self.guyspec()
    for c in self.commands:
      if c == "quit":
        self.stuff["quit"] = True
      elif isinstance(c, tuple):
        if c[0] == "move":
          moved = True
          guy.facing = c[1]
        elif c[0] == "select_tool":
          if TOOLS[c[1]] == "upgrade":
            self.upgrade()
          elif TOOLS[c[1]] != None:
            self.stuff["active_tool"] = TOOLS[c[1]]
        elif c[0] == "use_tool":
          cell = self.get(*c[1])
          tool = self.stuff["active_tool"]
          if tool \
          and isinstance(tool, type) \
          and issubclass(tool, objects.Thing) \
          and zone(*c[1]) == "middle" \
          and (
              cell["thing"] == None \
            or type(cell["thing"]) in (objects.Wall, objects.Barricade)
          ):
            if tool.cost <= self.stuff["energy"]:
              cell["thing"] = tool()
              self.stuff["energy"] -= tool.cost
            else:
              pass
              # TODO: Feedback here?
        pass
    if not moved:
      guy.facing = None
    self.commands = []

    # Next any turrets that are ready can fire (and we'll keep track of which
    # ones did so):
    attacks = []
    for x, col in enumerate(self.things):
      for y, cell in enumerate(col):
        thing = cell["thing"]
        if isinstance(thing, objects.Turret):
          thing.charge += 1
          if thing.charge >= thing.threshold:
            thing.charge = thing.threshold
            primary_target = None
            targets = set()
            multitarget = False
            target_type = lambda cell: isinstance(cell["thing"], objects.Enemy)
            if isinstance(thing, objects.TeslaTurret):
              multitarget = True
            if isinstance(thing, objects.MagnetTurret):
              multitarget = True
              target_type = lambda cell: cell["loot"] > 0
            for dist in range(1, thing.range + 1):
              for tx in range(x - dist, x + dist + 1):
                for ty in range(y - dist, y + dist + 1):
                  if zone(tx, ty) != "outside":
                    considering = self.get(tx, ty)
                    if target_type(considering):
                      if not primary_target:
                        primary_target = (tx, ty)
                      targets.add((tx, ty))
                      if not multitarget:
                        break
                if targets and not multitarget:
                  break
              if targets and not multitarget:
                break
            if targets:
              if isinstance(thing, objects.BurstTurret):
                for tx in (
                  primary_target[0] - 1,
                  primary_target[0],
                  primary_target[0] + 1
                ):
                  for ty in (
                    primary_target[1] - 1,
                    primary_target[1],
                    primary_target[1] + 1
                  ):
                    self.attack((tx, ty), thing.damage)
                attacks.append(
                  {
                    "source": (x, y),
                    "target": primary_target,
                    "type": thing.attack_type,
                  }
                )
                thing.charge = 0
              elif isinstance(thing, objects.TeslaTurret):
                for tx, ty in targets:
                  self.attack((tx, ty), thing.damage)
                  attacks.append(
                    {
                      "source": (x, y),
                      "target": (tx, ty),
                      "type": thing.attack_type,
                    }
                  )
                thing.charge = 0
              elif isinstance(thing, objects.MagnetTurret):
                siphoned = set()
                for tx, ty in targets:
                  cell = self.get(tx, ty)
                  while cell["loot"] > 0 \
                  and self.stuff["energy"] < MAX_ENERGY\
                  and thing.charge > 0:
                    cell["loot"] -= 1
                    self.stuff["energy"] += 1
                    thing.charge -= 1
                    siphoned.add((tx, ty))
                for s in siphoned:
                  attacks.append(
                    {
                      "source": (x, y),
                      "target": s,
                      "type": thing.attack_type,
                    }
                  )
              else:
                self.attack(primary_target, thing.damage)
                attacks.append(
                  {
                    "source": (x, y),
                    "target": (tx, ty),
                    "type": thing.attack_type,
                  }
                )
                thing.charge = 0

    # Next move the guy:
    pos, cell, guy = self.guyspec()
    next = guy.next_pos(self, *pos)
    if next and next != pos \
    and 0 <= next[0] < SIZE[0] \
    and 0 <= next[1] < SIZE[1]:
      target = self.things[next[0]][next[1]]
      if target["thing"] == None:
        target["thing"] = guy
        cell["thing"] = None
        self.stuff["guy_pos"] = next
      elif isinstance(target["thing"], objects.Enemy):
        damage = self.attack(next, self.stuff["g_sap_rate"], drop_loot=False)
        while damage > 0 \
        and self.stuff["g_carrying"] <= self.stuff["g_capacity"]:
          damage -= 1
          self.stuff["g_carrying"] += 1
        target["loot"] += damage
    pos = self.stuff["guy_pos"]
    # Loot near the guy gets collected:
    for x in (pos[0] - 1, pos[0], pos[0] + 1):
      for y in (pos[1] - 1, pos[1], pos[1] + 1):
        if (0 <= x < SIZE[0]) and (0 <= y < SIZE[1]):
          cell = self.things[x][y]
          while cell["loot"] > 0 \
          and self.stuff["g_carrying"] < self.stuff["g_capacity"]:
            cell["loot"] -= 1
            self.stuff["g_carrying"] += 1
    # When near base, loot gets converted to energy:
    if zone(*pos) == "close":
      while self.stuff["g_carrying"] > 0 and self.stuff["energy"] < MAX_ENERGY:
        self.stuff["g_carrying"] -= 1
        self.stuff["energy"] += 1

    # Next, move all of the enemies, if it's their turn to move:
    if self.stuff["phase"] % ENEMY_PHASE == 0:
      for x, col in enumerate(self.things):
        for y, cell in enumerate(col):
          thing = cell["thing"]
          if isinstance(thing, objects.Enemy):
            thing.moved = False
      for x, col in enumerate(self.things):
        for y, cell in enumerate(col):
          thing = cell["thing"]
          if isinstance(thing, objects.Enemy) and thing.moved == False:
            thing.moved = True
            next = thing.next_pos(self, x, y)
            if next and next != (x, y) and zone(*next) != "outside":
              target = self.things[next[0]][next[1]]
              if target["thing"] == None:
                target["thing"] = thing
                cell["thing"] = None
              elif not isinstance(target["thing"], objects.Enemy):
                if isinstance(thing, objects.BombBot) \
                and not isinstance(target["thing"], objects.Guy):
                  self.detonate(next, thing.damage)
                else:
                  self.attack(next, thing.damage)

    # Finally, decrement the wave counter for the next wave and spawn it if
    # it's time to do so:
    inbound = self.stuff["waves_inbound"]
    if len(inbound) < 2:
      inbound.append(waves.get(inbound[-1]["number"] + 1))
    inbound[0]["timer"] -= 1
    if inbound[0]["timer"] == 0:
      self.stuff["current_wave"] = inbound[0]["number"]
      arriving = inbound[0]
      self.stuff["waves_inbound"] = self.stuff["waves_inbound"][1:]
      leftovers = None
      for i, enemy in enumerate(arriving["contents"]):
        available = []
        for e in EDGES:
          if self.get(*e)["thing"] == None:
            available.append(e)
        if not available:
          leftovers = arriving["contents"][0:]
          break
        pos = None
        while (not pos) or self.get(*pos)["thing"] != None:
          pos = random.choice(available)
        heartwards = None
        if pos[0] == 0:
          heartwards = "east"
        elif pos[0] == SIZE[0] - 1:
          heartwards = "west"
        elif pos[1] == SIZE[1] - 1:
          heartwards = "north"
        elif pos[1] == 0:
          heartwards = "south"
        self.things[pos[0]][pos[1]]["thing"] = enemy(heartwards)
      if leftovers:
        inbound = [
          {
            "number": arriving["number"] + 0.01,
            "timer": 1,
            "contents": leftovers,
          }
        ] + inbound

    for d in self.stuff["detonations"]:
      attacks.append({"type": "detonation", "source": d, "target": d})
    self.stuff["detonations"] = []
    return attacks
