"""
render.py

Draws a State onto the screen.
"""

import os

import pygame

from engine import TOOLS, zone

def loadres(name):
  return pygame.image.load(os.path.join("res", name + ".png"))

floor = loadres("floor")
nobuild = loadres("floor-nobuild")
home = loadres("floor-home")
panel = loadres("panel-alt")
select = loadres("select")
selected = loadres("selected")
some_loot = loadres("smallloot")
much_loot = loadres("medloot")
so_much_loot = loadres("bigloot")

SCREENS = {
  "title": loadres("title"),
  "last_human": loadres("last_human"),
  "humanity_is_gone": loadres("humanity_is_gone"),
  "controls": loadres("howto-controls"),
  "pieces": loadres("howto-pieces"),
  "buildings": loadres("howto-buildings"),
  "howto": loadres("howto-instructions"),
  "get_ready": loadres("get_ready"),
  "game_over": loadres("game_over"),
}

font = pygame.font.Font(None, 20)
smallfont = pygame.font.Font(None, 16)
greenfont = pygame.font.Font(os.path.join("res", "BPdotsUnicaseLight.otf"), 32)

SOME_LOOT = 3
MUCH_LOOT = 10

BGCOLOR = (0, 0, 0)

TEXT_COLOR = (255, 255, 255)

def show_screen(surface, screen_name):
  """
  Shows a particular screen.
  """
  surface.fill(BGCOLOR)
  surface.blit(SCREENS[screen_name], (0, 0))
  pygame.display.flip()

def show_high_score(surface, wave):
  """
  Shows the game over/high score screen.
  """
  surface.fill(BGCOLOR)
  surface.blit(SCREENS["game_over"], (0, 0))
  hstext = greenfont.render(
    "YOU WERE DESTROYED BY WAVE " + str(int(wave)),
    True,
    (13, 221, 38)
  );
  tx = 380 - (hstext.get_width() // 2)
  ty = 400 - (hstext.get_height() // 2)
  surface.blit(hstext, (tx, ty))
  pygame.display.flip()

def charge_bar(
  surface,
  base,
  max,
  current,
  fg=(255, 255, 0),
  mg=(108, 108, 96),
  bg=(0, 0, 0)
):
  """
  Draws a charge bar given the upper-left of a 32x32 pixel space to draw it in
  and the max and current charge values. fg, mg, and bg colors may also be
  given. Raises a ValueError if the max (or current) values are above 30.
  """
  if max > 30 or current > 30:
    raise ValueError("Too much charge!")
  sx = base[0] + 1
  sy = base[1] + 32 - 2
  bx = sx + 3
  bonus_max = 0
  bonus = 0
  if max > 15:
    bonus_max = max - 15
    max = 15
  if current > 15:
    bonus = current - 15
    current = 15
  pygame.draw.rect(surface, bg, (sx - 1, sy - 2*max + 1, 4, 2*max + 2))
  if bonus_max:
    pygame.draw.rect(
      surface,
      bg,
      (bx - 1, sy - 2*bonus_max + 1, 4, 2*bonus_max + 2)
    )
  for i in range(max):
    x = sx
    y = sy - i*2
    pygame.draw.line(surface, mg, (x, y), (x + 1, y))
    if i < bonus_max:
      x = bx
      pygame.draw.line(surface, mg, (x, y), (x + 1, y))
  for i in range(current):
    x = sx
    y = sy - i*2
    pygame.draw.line(surface, fg, (x, y), (x + 1, y))
    if i < bonus:
      x = bx
      pygame.draw.line(surface, fg, (x, y), (x + 1, y))

def health_bar(
  surface,
  base,
  max,
  current,
  fg=(255, 255, 255),
  mg=(96, 96, 96),
  bg=(0, 0, 0)
):
  """
  Draws a health bar given the upper-left of a 32x32 pixel space to draw it in
  and the max and current health values. fg, mg, and bg colors may also be
  given. Raises a ValueError if the max (or current) values are above 30.
  """
  if max > 30 or current > 30:
    raise ValueError("Too much health!")
  sx = base[0] + 32 - 3
  sy = base[1] + 32 - 2
  bx = sx - 3
  bonus_max = 0
  bonus = 0
  if max > 15:
    bonus_max = max - 15
    max = 15
  if current > 15:
    bonus = current - 15
    current = 15
  pygame.draw.rect(surface, bg, (sx - 1, sy - 2*max + 1, 4, 2*max + 2))
  if bonus_max:
    pygame.draw.rect(
      surface,
      bg,
      (bx - 1, sy - 2*bonus_max + 1, 4, 2*bonus_max + 2)
    )
  for i in range(max):
    x = sx
    y = sy - i*2
    pygame.draw.line(surface, mg, (x, y), (x + 1, y))
    if i < bonus_max:
      x = bx
      pygame.draw.line(surface, mg, (x, y), (x + 1, y))
  for i in range(current):
    x = sx
    y = sy - i*2
    pygame.draw.line(surface, fg, (x, y), (x + 1, y))
    if i < bonus:
      x = bx
      pygame.draw.line(surface, fg, (x, y), (x + 1, y))

def render(surface, state, attacks, hover, hover_tool, frame):
  """
  Draws the given state onto the given surface. It also draws stuff to render
  the attack animations. This gets called 4 times per game tick, with frame
  values set to 0, 1, 2, and 3.
  """
  surface.fill(BGCOLOR)
  surface.blit(panel, (21*32, 0))
  frame_clip = (32*frame, 0, 32, 32)
  for x in range(state.size[0]):
    for y in range(state.size[1]):
      cursor = (x*32, y*32)
      cell = state.get(x, y)
      if zone(x, y) == "middle":
        surface.blit(floor, cursor)
      elif zone(x, y) == "close":
        surface.blit(home, cursor)
      else:
        surface.blit(nobuild, cursor)
      if cell["loot"] and cell["loot"] > 0:
        if cell["loot"] <= SOME_LOOT:
          surface.blit(some_loot, cursor, area = frame_clip)
        elif cell["loot"] <= MUCH_LOOT:
          surface.blit(much_loot, cursor, area = frame_clip)
        else:
          surface.blit(so_much_loot, cursor, area = frame_clip)
      thing = cell["thing"]
      if thing:
        pic = thing.sprite.subsurface(frame_clip)
        facing = thing.facing
        if hasattr(thing, "spritefacing"):
          if thing.spritefacing == "west":
            pic = pygame.transform.flip(pic, True, False)
        elif thing.facing:
          if thing.facing == "north":
            pic = pygame.transform.rotate(pic, 90)
          elif thing.facing == "west":
            pic = pygame.transform.rotate(pic, 180)
          elif thing.facing == "south":
            pic = pygame.transform.rotate(pic, 270)
        surface.blit(pic, cursor)
        if hasattr(thing, "health") and thing.health != None:
          health_bar(surface, cursor, thing.maxhealth, thing.health)
        if hasattr(thing, "charge") and thing.charge != None:
          charge_bar(surface, cursor, thing.threshold, thing.charge)
      if (x, y) == state.stuff["guy_pos"]:
        health_bar(
          surface,
          cursor,
          state.stuff["g_capacity"],
          state.stuff["g_carrying"], 
          fg=(255, 255, 0),
          mg=(108, 108, 96),
        )
  for t in range(9):
    cursor = (21*32 + 16, (t*2 + 1)*32)
    if isinstance(TOOLS[t], type):
      surface.blit(TOOLS[t].sprite, cursor, area = (0, 0, 32, 32))
      if (state.stuff["active_tool"] == TOOLS[t]):
        surface.blit(selected, cursor, area = frame_clip)
      cnums = smallfont.render(str(TOOLS[t].cost), True, TEXT_COLOR)
      tx = cursor[0] + 32 + (32 - cnums.get_width()) // 2
      ty = cursor[1] + (32 - cnums.get_height()) // 2
      surface.blit(cnums, (tx, ty))
  if hover:
    surface.blit(
      select,
      (hover[0]*32, hover[1]*32),
      area = frame_clip
     )
  if hover_tool != None and TOOLS[hover_tool] != None:
    surface.blit(
      select,
      (21*32 + 16, (hover_tool*2 + 1)*32),
      area = frame_clip
    )
  enums = font.render(str(state.stuff["energy"]), True, TEXT_COLOR)
  tx = 22 * 32 + (32 - enums.get_width()) // 2
  ty = 17 * 32 + (32 - enums.get_height()) // 2
  surface.blit(enums, (tx, ty))
  for a in attacks:
    color = (255, 255, 255)
    if a["type"] == "cucumber beam":
      cout = (181, 230, 29)
      cin = (153, 217, 234)
    elif a["type"] == "pea soup beam":
      cout = (0, 83, 166)
      cin = (28, 221, 33)
    elif a["type"] in ("detonation", "orange burst beam"):
      cout = (255, 242, 0)
      cin = (255, 127, 39)
    elif a["type"] == "tesla beam":
      cout = (232, 232, 255)
      cin = (255, 255, 255)
    elif a["type"] == "magnet beam":
      color = (255, 255, 0)
    sx, sy = a["source"]
    tx, ty = a["target"]
    ns = ty - sy
    ew = tx - sx
    if abs(ns) > abs(ew) + 1:
      orientation = "updown"
    elif abs(ew) > abs(ns) + 1:
      orientation = "leftright"
    else:
      orientation = "diagonal"
    sx = sx * 32 + 16
    sy = sy * 32 + 16
    tx = tx * 32 + 16
    ty = ty * 32 + 16
    starts = [[sx, sy], [sx, sy], [sx, sy]]
    ends = [[tx, ty], [tx, ty], [tx, ty]]
    if orientation in ("updown", "diagonal"):
      starts[0][0] -= 1
      starts[2][0] += 1
      ends[0][0] -= 1
      ends[2][0] += 1
    if orientation in ("leftright", "diagonal"):
      if ns > 0:
        if ew > 0:
          starts[0][1] += 1
          starts[2][1] -= 1
          ends[0][1] += 1
          ends[2][1] -= 1
        elif ew < 0:
          starts[0][1] -= 1
          starts[2][1] += 1
          ends[0][1] -= 1
          ends[2][1] += 1
      elif ns < 0:
        if ew > 0:
          starts[0][1] -= 1
          starts[2][1] += 1
          ends[0][1] -= 1
          ends[2][1] += 1
        elif ew < 0:
          starts[0][1] += 1
          starts[2][1] -= 1
          ends[0][1] += 1
          ends[2][1] -= 1
      else:
        starts[0][1] -= 1
        starts[2][1] += 1
        ends[0][1] -= 1
        ends[2][1] += 1
    if a["type"] == "magnet beam":
      pygame.draw.line(surface, color, starts[1], ends[1])
    else:
      pygame.draw.line(surface, cout, starts[0], ends[0])
      pygame.draw.line(surface, cin, starts[1], ends[1])
      pygame.draw.line(surface, cout, starts[2], ends[2])
      pygame.draw.circle(surface, cin, starts[1], 3)
      pygame.draw.circle(surface, cin, ends[1], 4)
      pygame.draw.circle(surface, cout, ends[1], 3)
      pygame.draw.circle(surface, cout, starts[1], 4)
      if a["type"] in ("detonation", "orange burst beam"):
        pygame.draw.circle(surface, cin, ends[1], 32, 2)
        pygame.draw.circle(surface, cout, ends[1], 34, 2)
  pygame.display.flip()
